/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "fmsound.h"

#include "exception.h"

//
// FMSoundManager
//
void FMSoundManager::initialize() {
  if(FSOUND_GetVersion() < FMOD_VERSION)
    Exception::throws("Wrong FMOD DLL version: 3.5 needed");
  FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND);
  FSOUND_SetDriver(0);
  int channels = FSOUND_GetNumHardwareChannels();
  if(channels < MIN_HARDWARE_CHANNELS)
    FSOUND_SetMinHardwareChannels(channels = MIN_HARDWARE_CHANNELS);
  if(!FSOUND_Init(44100,MIN_HARDWARE_CHANNELS,0))
    Exception::throws("FMOD Sound Init Error");
  FSOUND_SetSFXMasterVolume(255);
}

int FMSoundManager::channels = 0;

//
// FMSample
//

FMSample::FMSample(int length, unsigned char* buffer) {
  init(length,buffer,FSOUND_2D);
}

FMSample::FMSample() {
}

void FMSample::init(int length, unsigned char* buffer, int flags) {
  sample = FSOUND_Sample_Load(
    FSOUND_FREE,buffer,flags|FSOUND_LOADMEMORY,length
  );
}

FMSample::~FMSample() {
  FSOUND_Sample_Free(sample);
}

FMSound* FMSample::createSound() {
  return new FMSound(FMSound::IS_CHANNEL_LOST,this);
}

//
// FM3DSample
//

FM3DSample::FM3DSample(int length, unsigned char* buffer): invMinDistance(1) {
  init(length,buffer,FSOUND_HW3D);
}

FM3DSound* FM3DSample::create3DSound() {
  return new FM3DSound(FMSound::IS_CHANNEL_LOST,this);
}

//
// FMSound
//

FMSound::FMSound(int c, FMSample* s): channel(c), sample(s) {
}

FMSound::~FMSound() {
  stop();
}

FMSound::SoundStatus FMSound::getSoundStatus() {
  if(isAudible()) {
    FSOUND_SAMPLE* s = FSOUND_GetCurrentSample(channel);
    if(s == sample->getPointer()) {
      if(FSOUND_IsPlaying(channel) == TRUE)
        return IS_SOUND_PLAYING;
      else
        channel = WAS_SOUND_STOPPED;
    } else {
      channel = sample->isLooping()? IS_CHANNEL_LOST: WAS_SOUND_STOPPED;
    }
  }
  return (SoundStatus)channel;
}

//
// FM3DSound
//

FM3DSound::FM3DSound(int c, FM3DSample* s): FMSound(c,s) {
}

//
// FMMusic
//

FMMusic::FMMusic(int length, unsigned char* buffer) {
  module = FMUSIC_LoadSongMemory(buffer,length);
}

FMMusic::~FMMusic() {
  FMUSIC_FreeSong(module);
}
