#ifndef GLCAMERA_H
#define GLCAMERA_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "math3d.h"

//
// GLFrustum
//
class GLFrustum {
protected:
  float clipPlaneNear;
  float clipPlaneFar;
  float frustum[6][4];
public:
  GLFrustum();
  float getFrustum(int plane, int coo) {return frustum[plane][coo];}
  float getClipPlaneNear() {return clipPlaneNear;}
  void update();
};

//
// GLNearestLight
//
class GLNearestLight {
public:
  bool isDirectional;
  M3Vector vector;
};

//
// GLCamera
//
class GLCamera: public M3Reference, public GLFrustum {
private:
  M3Vector viewXY; // for GLBillboard
  float fieldOfView;
  bool renderingReflections;
  GLNearestLight nearestLight;
  M3Vector lastApplyPosition;
public:
  GLCamera();
  static void setOrtho();
  static void saveView();
  static void restoreView();
  void setPerspective(float fov, float aspect, float zn, float zf);
  void updatePerspective(float aspect)
    {setPerspective(fieldOfView,aspect,clipPlaneNear,clipPlaneFar);}
  void setRenderingReflections(bool rr) {renderingReflections = rr;}
  bool isRenderingReflections() {return renderingReflections;}
  void setNearestLight(bool isDir, const M3Vector& vec)
    {nearestLight.isDirectional = isDir; nearestLight.vector = vec;}
  bool isNearestLightDirectional() {return nearestLight.isDirectional;}
  M3Vector& getNearestLightVector() {return nearestLight.vector;}
  M3Vector& getViewXY() {return viewXY;}
  void setVelocityToZero()
    {lastApplyPosition.set(getPositionX(),getPositionY(),-getPositionZ());}
  void applyRot();
  void apply();
};

#endif // GLCAMERA_H
