#ifndef GLCONSOLE_H
#define GLCONSOLE_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "slinterpreter.h"
#include "datasets.h"

//
// GlConsole
//
class GLConsole {
public:
  enum {
    CONSOLE_ROWS = 30,
    CONSOLE_COLS = 60,
  };
private:
  static GLConsole* console;
  SLInterpreter scriptInterpreter;
  DSArray<SLInterpreter> interpreters;
  bool suspended;
  char screen[CONSOLE_ROWS][CONSOLE_COLS+1];
  int currentRow;
  int currentCol;
protected:
  void drawCursor() {
    screen[currentRow][currentCol] = '_';
    screen[currentRow][currentCol+1] = '\0';
  }
  void newLine() {
    screen[currentRow][currentCol] = '\0';
    if(++currentRow >= CONSOLE_ROWS) currentRow = 0;
    currentCol = 0;
  }
public:
  GLConsole();
  static GLConsole& get() {return *console;}
  int addInterpreter();
  void removeInterpreter(int idx) {interpreters.removeElement(0);}
  SLInterpreter* getInterpreter(int idx) {return interpreters.getElement(idx);}
  SLInterpreter& getScriptInterpreter() {return scriptInterpreter;}
  bool isSuspended() {return suspended;}
  void continueExecution(float deltaT);
  void suspend();
  void resume();
  int getRows() {return CONSOLE_ROWS;}
  int getCols() {return CONSOLE_COLS;}
  char* getRow(int idx);
  void print(char ch);
  void print(char* text);
  void println(char* text) {print(text); print('\n');}
  void printPrompt() {if(currentCol != 0) print('\n'); print("> ");}
  void keyPressed(char ch);
  bool executeCommand(char* commandLine);
};

#endif // GLCONSOLE_H

