/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glmaterial.h"

//
// GLMaterial
//

GLMaterial::GLMaterial():
  ambient(0.2f,0.2f,0.2f), diffuse(0.8f,0.8f,0.8f),
  specular(0.0f,0.0f,0.0f), emissive(0.0f,0.0f,0.0f),
  shininess(10.0f),
  diffuseTexture(NULL), glossTexture(NULL),
  environmentTexture(NULL), environment(-1.0f)
{}

void GLMaterial::destroyObject() {
  if(environmentTexture) environmentTexture->releaseReference();
  if(diffuseTexture) diffuseTexture->releaseReference();
  if(glossTexture) glossTexture->releaseReference();
  delete this;
}

void GLMaterial::applyColor() {
  glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,shininess);
  glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,ambient.getFloatArray());
  glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,diffuse.getFloatArray());
  glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,specular.getFloatArray());
  glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,emissive.getFloatArray());
}

void GLMaterial::applyEnvironment() {
  float env[] = {1,1,1,environment};
  glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,env);
  glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,env);
  glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,env);
  glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,env);
  environmentTexture->apply();
}

//
// GLBumpedMaterial
//

GLBumpedMaterial::GLBumpedMaterial(): bumpTexture(NULL), bump(0.0f) {
}

void GLBumpedMaterial::destroyObject() {
  if(bumpTexture) bumpTexture->releaseReference();
  GLMaterial::destroyObject();
}
