/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "opengl.h"
#include "globject.h"

#include "datasets.cpp"

//
// GLClip
//
GLClip::GLClip(): maxRadius(0), clipped(false), distance(0) {
}

//
// GLObject
//

GLObject::GLObject(): visible(false), transparent(false) {
}

GLObject::~GLObject() {
}

void GLObject::clip(GLCamera& camera, M3Matrix* parentTransform) {
  clipped = false;
  M3Vector pos(getPositionX(),getPositionY(),getPositionZ());
  if(parentTransform) parentTransform->multiply(pos);
  for(int ct = 0; ct < 5; ct++) {
    float dist =
      camera.getFrustum(ct,0)*pos.getX()+
      camera.getFrustum(ct,1)*pos.getY()+
      camera.getFrustum(ct,2)*pos.getZ()+
      camera.getFrustum(ct,3)
    ;
    if(dist <= -getMaxRadius()) {
      clipped = true;
      break;
    }
  }
  distance =
    camera.getFrustum(5,0)*pos.getX()+
    camera.getFrustum(5,1)*pos.getY()+
    camera.getFrustum(5,2)*pos.getZ()+
    camera.getFrustum(5,3)
  ;
  if(distance <= -getMaxRadius()) clipped = true;
  if(camera.isRenderingReflections() && clipped) {
    clipped = false;
    pos.setY(-pos.getY());
    for(int ct = 0; ct < 5; ct++) {
      float dist =
        camera.getFrustum(ct,0)*pos.getX()+
        camera.getFrustum(ct,1)*pos.getY()+
        camera.getFrustum(ct,2)*pos.getZ()+
        camera.getFrustum(ct,3)
      ;
      if(dist <= -getMaxRadius()) {
        clipped = true;
        break;
      }
    }
    distance =
      camera.getFrustum(5,0)*pos.getX()+
      camera.getFrustum(5,1)*pos.getY()+
      camera.getFrustum(5,2)*pos.getZ()+
      camera.getFrustum(5,3)
    ;
    if(distance <= -getMaxRadius()) clipped = true;
  }
  distance = abs(distance);
}

//
// GLObjects
//
void GLObjects::clip(GLCamera& camera, M3Matrix* parentTransform) {
  GLObject::clip(camera,parentTransform);
  if(!isClipped()) {
    GLObject* obj = getFirstObject();
    if(parentTransform) {
      while(obj) {
        M3Matrix groupTransform(*parentTransform);
        groupTransform.multiply(*parentTransform);
        obj->clip(camera,&groupTransform);
        obj = getNextObject();
      }
    } else {
      while(obj) {
        obj->clip(camera,(M3Matrix*)this);
        obj = getNextObject();
      }
    }
  }
}

void GLObjects::render(GLCamera& camera) {
	glPushMatrix();
	glMultMatrixf(getTransform());
  GLObject* obj = getFirstObject();
  while(obj) {
    if((!obj->isClipped()) && obj->isVisible())
      obj->render(camera);
    obj = getNextObject();
  }
	glPopMatrix();
}
