#ifndef GLOBJECT_H
#define GLOBJECT_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "math3d.h"
#include "glcamera.h"
#include "datasets.h"

//
// GLClip
//
class GLClip {
protected:
  bool clipped;
  float distance;
  float maxRadius;
public:
  GLClip();
  void setMaxRadius(float rd) {maxRadius = rd;}
  float getMaxRadius() {return maxRadius;}
  float getDistance() {return distance;}
  bool isClipped() {return clipped;}
  virtual void clip(GLCamera& camera, M3Matrix* parentTransform = NULL) = 0;
};

//
// GLObject
//
class GLObject: public GLClip, public M3Reference, public DSSortableObject {
protected:
  bool visible;
  bool transparent;
public:
  GLObject();
  virtual ~GLObject();
  void setVisible(bool v) {visible = v;}
  bool isVisible() {return visible;}
  void setTransparent(bool t) {transparent = t;}
  bool isTransparent() {return transparent;}
  bool bounds(const M3Vector& vec) {
    float dx = getPositionX()-vec.getX();
    float dy = getPositionY()-vec.getY();
    float dz = getPositionZ()-vec.getZ();
    return dx*dx+dy*dy+dz*dz <= maxRadius*maxRadius;
  }
  float getSortIndex() {return getDistance();}
  virtual void clip(GLCamera& camera, M3Matrix* parentTransform = NULL);
  virtual void render(GLCamera& camera) = 0;
};

//
// GLObjects
//
class GLObjects: public GLObject {
private:
  DSList<GLObject> objects;
public:
  void addObject(GLObject* obj) {
    obj->setVisible(true); objects.addElement(obj);
    if(getMaxRadius() < obj->getMaxRadius()) setMaxRadius(obj->getMaxRadius());
  }
  GLObject* getFirstObject() {return objects.resetToFirstElement();}
  GLObject* getNextObject() {return objects.nextElement();}
  int getObjectsCount() {return objects.getLength();}
  void clip(GLCamera& camera, M3Matrix* parentTransform = NULL);
  void render(GLCamera& camera);
};

#endif // GLOBJECT_H
