#ifndef GLTERRAIN_H
#define GLTERRAIN_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glmaterial.h"

#include <math.h> // for sqrt()

//
// GLFlatTerrain
//
class GLFlatTerrain {
private:
  unsigned int list;
  GLMaterial* material;
  bool reflective;
  bool visible;
  float size;
  int tiles;
protected:
  enum {SIDE_QUADS_COUNT = 32};
public:
  GLFlatTerrain(GLMaterial* mat, float siz, int tils = 1);
  ~GLFlatTerrain();
  GLMaterial& getMaterial() {return *material;}
  void setVisible(bool v) {visible = v;}
  bool isVisible() {return visible;}
  void setReflective(bool r) {reflective = r;}
  bool isReflective() {return reflective;}
  virtual void render(GLCamera& camera);
};

//
// GLHeightField
//
class GLHeightField: public GLObject {
private:
  int listCount;
  unsigned int list;
  GLMaterial* material;
  float width, depth, height, waterLevel;
  int detailedTiles;
  DRImage* heightMap;
protected:
  enum LodMode {
    NEAR_LOD = 0,
    NORMAL_LOD,
    FAR_LOD,
    FARTEST_LOD,
    LODS_COUNT
  };
public:
  GLHeightField(
    DRImage* hm, GLMaterial* mat, float w, float d, float h,
    float wl = 0, int detTils = 1
  );
  ~GLHeightField();
  GLMaterial& getMaterial() {return *material;}
  void setMaterial(GLMaterial* mat) {
    if(material) material->releaseReference();
    material = mat->acquireReference();
  }
  void setHeightMap(DRImage* hm);
  void setSize(float w, float d, float h, float wl) {
    width = w; depth = d; height = h; waterLevel = wl;
    setMaxRadius(sqrt(w*w+d*d+h*h)*0.5f);
  }
  float getHeightAtAbsolute(float x, float y);
  float getHeightAtRelative(float x, float y);
  void compile();
  void render(GLCamera& camera);
};

#endif // GLTERRAIN_H