#ifndef GLWORLD_H
#define GLWORLD_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "globject.h"
#include "glbackground.h"
#include "glterrain.h"
#include "gllight.h"
#include "glsound.h"

//
// GLWorld
//
class GLWorld {
private:
  DSArray<GLBackground> backgrounds;
  DSSortedArray<GLObject> opaqueObjects;
  DSSortedArray<GLObject> transparentObjects;
  DSSortedArray<GLSoundSource> soundSources;
  DSSortedArray<GLObject> lights;
  GLFlatTerrain* flatTerrain;
  GLSunLight* sun;
  GLColor ambient;
public:
  GLWorld();
  ~GLWorld();
  int getOpaqueObjectsCount() {return opaqueObjects.getSize();}
  int getTransparentObjectsCount() {return transparentObjects.getSize();}
  GLObject* getOpaqueObject(int idx)
    {return opaqueObjects.getElement(idx);}
  GLObject* getTransparentObject(int idx)
    {return transparentObjects.getElement(idx);}
  void addObject(GLObject* object) {
    (object->isTransparent()? transparentObjects: opaqueObjects).addElement(
      (DSSortableObject*)object
    );
  }
  bool removeOpaqueObject(GLObject* o) {return opaqueObjects.removeElement(o);}
  bool removeTransparentObject(GLObject* o)
     {return transparentObjects.removeElement(o);}
  void removeObject(GLObject* o)
    {removeOpaqueObject(o) && removeTransparentObject(o);}
  void deleteObjects() {
    transparentObjects.deleteElements();
    opaqueObjects.deleteElements();
    deleteSoundSources();
  }
  int getSoundSourcesCount() {return soundSources.getSize();}
  GLSoundSource* getSoundSource(int idx) {return soundSources.getElement(idx);}
  void addSoundSource(GLSoundSource* soundSource)
    {soundSources.addElement((DSSortableObject*)soundSource);}
  void deleteSoundSources() {soundSources.deleteElements();}
  int getLightsCount() {return lights.getSize();}
  GLLight* getLight(int idx) {return (GLLight*)lights.getElement(idx);}
  void addLight(GLLight* light) {lights.addElement((DSSortableObject*)light);}
  void deleteLights() {lights.deleteElements();}
  void removeEnvironment();
  void empty() {deleteObjects(); deleteLights(); removeEnvironment();}
  void setBackground(GLBackground* bg);
  void addBackground(GLBackground* bg) {backgrounds.addElement(bg);}
  void removeBackground(int idx) {backgrounds.removeElement(idx);}
  void removeBackground(GLBackground* b) {backgrounds.removeElement(b);}
  void setFlatTerrain(GLFlatTerrain* t);
  void setSun(GLSunLight* s);
  void setAmbient(float r, float g, float b) {ambient.set(r,g,b);}
  void disableFog() {glDisable(GL_FOG);}
  void setFogColor(GLColor& fogColor)
    {glFogfv(GL_FOG_COLOR,fogColor.getFloatArray());}
  void enableFog(float dist, GLColor& fogColor);
  void render(GLCamera& camera);
};

#endif // GLWORLD_H
