/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "math3d.h"

//
// M3Vector
//

M3Vector::M3Vector(M3Float x, M3Float y, M3Float z) {
  set(x,y,z);
}

M3Vector::M3Vector(M3Float* v) {
  set(v);
}

M3Vector::M3Vector(M3Vector& v) {
  set(v);
}

//
// M3Matrix
//
M3Matrix::M3Matrix(M3Matrix& m) {set(m);}

M3Matrix::M3Matrix(
  M3Float m11, M3Float m12, M3Float m13, M3Float m14,
  M3Float m21, M3Float m22, M3Float m23, M3Float m24,
  M3Float m31, M3Float m32, M3Float m33, M3Float m34,
  M3Float m41, M3Float m42, M3Float m43, M3Float m44
) {
  set(m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44);
}

M3Matrix::M3Matrix(M3Vector& v1, M3Vector &v2, M3Vector &v3) {
  set(
    v1.getX(), v2.getX(), v3.getX(), 0,
    v1.getY(), v2.getY(), v3.getY(), 0,
    v1.getZ(), v2.getZ(), v3.getZ(), 0,
    0, 0, 0, 1
  );
}

//
// M3RotationX
//
M3RotationX::M3RotationX(M3Float angle):
  M3Matrix(1,0,0,0,  0,cos(angle),sin(angle),0,  0,-sin(angle),cos(angle),0)
{}

//
// M3RotationY
//
M3RotationY::M3RotationY(M3Float angle):
  M3Matrix(cos(angle),0,-sin(angle),0,  0,1,0,0,  sin(angle),0,cos(angle),0)
{}

//
// M3RotationZ
//
M3RotationZ::M3RotationZ(M3Float angle):
  M3Matrix(cos(angle),sin(angle),0,0,  -sin(angle),cos(angle),0,0,  0,0,1,0)
{}

//
// M3Reference
//
void M3Reference::setTransform(
  M3Float x, M3Float y, M3Float z, M3Float theta, M3Float phi, M3Float rho
) {
  setPosition(x,y,z);
  float cT = cos(theta);
  float sT = sin(theta);
  float cP = cos(phi);
  float sP = sin(phi);
  float cR = cos(rho);
  float sR = sin(rho);
  setSideDirection(cP*cR,cP*sR,-sP);
  float sTsP = sT*sP;
  setUpDirection(sTsP*cR-cT*sR,sTsP*sR+cT*cR,sT*cP);
  float cTsP = cT*cP;
  setViewDirection(cTsP*cR+sT*sR,cTsP*sR-sT*cR,cT*cP);
}

