/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "phsimulator.h"

//
// PHSimulator
//
const M3Float PHSimulator::DELTA_T = 1.0f/PHSimulator::INTEGRATIONS_PER_SECOND;

PHSimulator::PHSimulator():
  positions0(32), positions1(32), accelerations(32),
  bodies(16), positions(&positions0), oldPositions(&positions1)
{}

void PHSimulator::runTime(M3Float timeStep) {
  int steps = (int)(timeStep*(1/DELTA_T)+0.5f);
  for(int ct = steps; ct > 0; ct--)
    integrate();
  for(int ct = 0; ct < getBodiesCount(); ct++)
    getBody(ct)->updateReference();
}

void PHSimulator::integrate() {
  for(int ct = 0; ct < accelerations.getSize(); ct++) {
    M3Vector& pos = getPosition(ct);
    M3Vector& oldPos = getOldPosition(ct);
    oldPos.set(
      pos.getX()+pos.getX()-oldPos.getX(),
      pos.getY()+pos.getY()-oldPos.getY()-(9.81/2*DELTA_T*DELTA_T),
      pos.getZ()+pos.getZ()-oldPos.getZ()
    );
    if(hasAcceleration(ct)) {
      M3Vector& acc = getAcceleration(ct);
      oldPos.add(acc.scale(DELTA_T*DELTA_T));
      zeroAcceleration(ct);
    }
  }
  DSStaticArray<M3Vector>* tmp = positions;
  positions = oldPositions;
  oldPositions = tmp;
}

//
// PHBody
//
PHBody::PHBody(PHSimulator& sim, PHEnvironment& env, M3Reference& ref):
  simulator(sim), environment(env), reference(ref), mass(0)
{
  simulator.addBody(this);
  positionIndex = simulator.addPosition();
  orientationIndex = simulator.addOrientation();
}

void PHBody::updateReference() {
  M3Vector& pos = simulator.getPosition(positionIndex);
  M3Vector& orient = simulator.getOrientation(orientationIndex);
  M3Vector omega = orient;
  omega.subtract(lastStepOrientation);
  reference.setTransform(pos,omega);
  lastStepOrientation = orient;
}
