#ifndef SLINTERPRETER_H
#define SLINTERPRETER_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "small/amx.h"

#include "dataread.h"

//
// SLInterpreter
//

class SLInterpreter {
private:
  unsigned char* program;
  float cpuFrequency;
  cell returnValue;
  bool running;
  AMX amx;
protected:
  int registerNatives();
public:
  SLInterpreter();
  ~SLInterpreter();
  bool load(DRData* prog);
  bool load(char* fileName);
  int initMain() {
    amx_timer = 0;
    int ret = amx_Exec(&amx,&returnValue,AMX_EXEC_MAIN,0);
    running = (ret == AMX_ERR_SLEEP);
    return ret;
  }
  int execMain(float deltaT) {
    amx_timer = (int)(deltaT*cpuFrequency+0.5);
    int ret = amx_Exec(&amx,&returnValue,AMX_EXEC_CONT,0);
    running = (ret == AMX_ERR_SLEEP);
    return ret;
  }
  void setCPUFrequency(float f) {cpuFrequency = f;}
  float getCPUFrequency() {return cpuFrequency;}
  void setRunning(bool r) {running = false;}
  bool isRunning() {return running;}
  cell getReturnValue() {return returnValue;}
};

#endif // SLINTERPRETER_H
